// file      : xsd/cxx/tree/parser-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/parser-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct Element: Traversal::Element,
                      GlobalElementBase,
                      protected virtual Context
      {
        Element (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!doc_root_p (e))
            return;

          String const& name (ename (e));
          String type (type_name (e));
          String properties_type (xs_ns_name () + L"::properties");
          String error_handler (xs_ns_name () + L"::error_handler");

          // Note that I am using fq-name in function calls because g++ gets
          // confused if the name is 'type'. (see tests/schema/anonymous)
          //

          // URI.
          //
          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const " << string_type << "& u," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0," << endl
             << "(f & " << flags_type << "::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::parse< " << char_type << " > (" <<
            "u, h, p, f));"
             << endl
             << "h.throw_if_failed< ::xsd::cxx::tree::parsing< " <<
            char_type << " > > ();"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "d.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "if (f & " << flags_type << "::keep_dom)" << endl
             << "d.release ();"
             << endl
             << "return r;"
             << "}";


          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const " << string_type << "& u," << endl
             << error_handler << "& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0," << endl
             << "(f & " << flags_type << "::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::parse< " << char_type << " > (" <<
            "u, h, p, f));"
             << endl
             << "if (!d)" << endl
             << "throw ::xsd::cxx::tree::parsing< " << char_type << " > ();"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "d.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "if (f & " << flags_type << "::keep_dom)" << endl
             << "d.release ();"
             << endl
             << "return r;"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const " << string_type << "& u," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::parse< " << char_type << " > (" <<
            "u, h, p, f));"
             << endl
             << "if (!d)" << endl
             << "throw ::xsd::cxx::tree::parsing< " << char_type << " > ();"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "d.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "if (f & " << flags_type << "::keep_dom)" << endl
             << "d.release ();"
             << endl
             << "return r;"
             << "}";


          // istream
          //
          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::std::istream& is," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0," << endl
             << "(f & " << flags_type << "::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is);"
             << "::xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::std::istream& is," << endl
             << error_handler << "& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0," << endl
             << "(f & " << flags_type << "::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is);"
             << "::xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::std::istream& is," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::sax::std_input_source isrc (is);"
             << "::xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::std::istream& is," << endl
             << "const " << string_type << "& sid," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0," << endl
             << "(f & " << flags_type << "::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is, sid);"
             << "::xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::std::istream& is," << endl
             << "const " << string_type << "& sid," << endl
             << error_handler << "& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0," << endl
             << "(f & " << flags_type << "::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is, sid);"
             << "::xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::std::istream& is," << endl
             << "const " << string_type << "& sid," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::sax::std_input_source isrc (is, sid);"
             << "::xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";


          // InputSource.
          //
          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const ::xercesc::DOMInputSource& i," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::parse< " << char_type << " > (" <<
            "i, h, p, f));"
             << endl
             << "h.throw_if_failed< ::xsd::cxx::tree::parsing< " <<
            char_type << " > > ();"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "d.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "if (f & " << flags_type << "::keep_dom)" << endl
             << "d.release ();"
             << endl
             << "return r;"
             << "}";

          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const ::xercesc::DOMInputSource& i," << endl
             << error_handler << "& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::parse< " << char_type << " > (" <<
            "i, h, p, f));"
             << endl
             << "if (!d)" << endl
             << "throw ::xsd::cxx::tree::parsing< " << char_type << " > ();"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "d.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "if (f & " << flags_type << "::keep_dom)" << endl
             << "d.release ();"
             << endl
             << "return r;"
             << "}";


          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const ::xercesc::DOMInputSource& i," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::parse< " << char_type << " > (" <<
            "i, h, p, f));"
             << endl
             << "if (!d)" << endl
             << "throw ::xsd::cxx::tree::parsing< " << char_type << " > ();"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "d.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "if (f & " << flags_type << "::keep_dom)" << endl
             << "d.release ();"
             << endl
             << "return r;"
             << "}";


          // DOM.
          //

          Boolean fund (false);

          {
            IsFundamentalType t (fund);
            t.dispatch (e.type ());
          }

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          Boolean polymorphic (!fund &&
                               options.value<CLI::generate_polymorphic> () &&
                               !e.type ().context ().count ("anonymous"));

          // const DOMDocument&
          //
          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (const ::xercesc::DOMDocument& d," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "& p)"
             << "{"
             << "if (f & " << flags_type << "::keep_dom)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > " <<
            "c (" << endl
             << "static_cast< ::xercesc::DOMDocument* > (d.cloneNode (true)));"
             << endl
             << "::std::auto_ptr< " << type << " > r (" << endl
             << fq_name (e) << " (" << endl
             << "c.get (), f | " << flags_type << "::own_dom, p));"
             << endl
             << "c.release ();"
             << "return r;"
             << "}"
             << "const ::xercesc::DOMElement& e (*d.getDocumentElement ());"
             << "const " << qname_type << " n (" << endl
             << "::xsd::cxx::xml::dom::name< " << char_type << " > (e));"
             << endl;

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_factory_map_instance.
            //
            os << "::xsd::cxx::tree::type_factory_map< " << char_type <<
              " >& tfm (" << endl
               << "::xsd::cxx::tree::type_factory_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "::std::auto_ptr< " << any_type << " > tmp (" << endl
               << "tfm.create (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "&::xsd::cxx::tree::factory_impl< " << type << " >," << endl
               << "true, true, e, n, f, 0));"
               << endl
               << "if (tmp.get () != 0)"
               << "{"
               << "::std::auto_ptr< " << type << " > r (" << endl
               << "dynamic_cast< " << type << "* > (tmp.get ()));"
               << endl
               << "if (r.get ())" << endl
               << "tmp.release ();"
               << "else" << endl
               << "throw ::xsd::cxx::tree::not_derived< " << char_type <<
              " > ();"
               << endl;
          }
          else
          {
            os << "if (n.name () == " << L << "\"" << e.name () << "\"" <<
              " &&" << endl
               << "n.namespace_ () == " << L << "\"" << e.namespace_().name () << "\")"
               << "{";

            if (fund)
            {
              os << "::std::auto_ptr< " << type << " > r (" << endl
                 << "new " << type << " (" << endl
                 << "::xsd::cxx::tree::traits< " << type << ", " <<
                char_type << " >::create (" << endl
                 << "e, f, 0)));";
            }
            else
            {
              os << "::std::auto_ptr< " << type << " > r (" << endl
                 << "::xsd::cxx::tree::traits< " << type << ", " <<
                char_type << " >::create (" << endl
                 << "e, f, 0));";
            }
          }

          os << "return r;"
             << "}"
             << "throw ::xsd::cxx::tree::unexpected_element < " <<
            char_type << " > (" << endl
             << "n.name ()," << endl
             << "n.namespace_ ()," << endl
             << L << "\"" << e.name () << "\"," << endl
             << L << "\"" << e.namespace_().name () << "\");"
             << "}";


          // DOMDocument*
          //
          os << "::std::auto_ptr< " << type << " >" << endl
             << name << " (::xercesc::DOMDocument* d," << endl
             << flags_type << " f," << endl
             << "const " << properties_type << "&)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > " <<
            "c (" << endl
             << "((f & " << flags_type << "::keep_dom) &&" << endl
             << "!(f & " << flags_type << "::own_dom))" << endl
             << "? static_cast< ::xercesc::DOMDocument* > (" <<
            "d->cloneNode (true))" << endl
             << ": 0);"
             << endl
             << "const ::xercesc::DOMElement& e (" << endl
             << "c.get ()" << endl
             << "? *c->getDocumentElement ()" << endl
             << ": *d->getDocumentElement ());"
             << "const " << qname_type << " n (" << endl
             << "::xsd::cxx::xml::dom::name< " << char_type << " > (e));"
             << endl;

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_factory_map_instance.
            //
            os << "::xsd::cxx::tree::type_factory_map< " << char_type <<
              " >& tfm (" << endl
               << "::xsd::cxx::tree::type_factory_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "::std::auto_ptr< " << any_type << " > tmp (" << endl
               << "tfm.create (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "&::xsd::cxx::tree::factory_impl< " << type << " >," << endl
               << "true, true, e, n, f, 0));"
               << endl
               << "if (tmp.get () != 0)"
               << "{";
          }
          else
          {
            os << "if (n.name () == " << L << "\"" << e.name () << "\"" <<
              " &&" << endl
               << "n.namespace_ () == " << L << "\"" << e.namespace_().name () << "\")"
               << "{";

            if (fund)
            {
              os << "::std::auto_ptr< " << type << " > r (" << endl
                 << "new " << type << " (" << endl
                 << "::xsd::cxx::tree::traits< " << type << ", " <<
                char_type << " >::create (" << endl
                 << "e, f, 0)));";
            }
            else
            {
              os << "::std::auto_ptr< " << type << " > r (" << endl
                 << "::xsd::cxx::tree::traits< " << type << ", " <<
                char_type << " >::create (" << endl
                 << "e, f, 0));";
            }
          }

          os << "c.release ();";


          if (polymorphic)
          {
            os << endl
               << "::std::auto_ptr< " << type << " > r (" << endl
               << "dynamic_cast< " << type << "* > (tmp.get ()));"
               << endl
               << "if (r.get ())" << endl
               << "tmp.release ();"
               << "else" << endl
               << "throw ::xsd::cxx::tree::not_derived< " << char_type <<
              " > ();"
               << endl;
          }

          os << "return r;"
             << "}"
             << "throw ::xsd::cxx::tree::unexpected_element < " <<
            char_type << " > (" << endl
             << "n.name ()," << endl
             << "n.namespace_ ()," << endl
             << L << "\"" << e.name () << "\"," << endl
             << L << "\"" << e.namespace_().name () << "\");"
             << "}";
        }

      private:
        String
        type_name (Type& e)
        {
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (e.type ());

          return o.str ();
        }
      };
    }

    Void
    generate_parser_source (Context& ctx,
                            UnsignedLong first,
                            UnsignedLong last)
    {
      ctx.os << "#include <istream>" << endl
             << "#include <xercesc/framework/Wrapper4InputSource.hpp>" << endl
             << "#include <xsd/cxx/xml/sax/std-input-source.hxx>" << endl
             << "#include <xsd/cxx/tree/error-handler.hxx>" << endl
             << endl;

      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx, first, last);
      Element element (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names >> element;

      schema.dispatch (ctx.schema_root);
    }
  }
}
